
namespace GossamerVR.WearableAsset
{
    using System;
    using System.Collections;
    using System.Collections.Generic;   
    using UnityEngine;
    using UnityEngine.UI;
    using SimpleJSON;

    using Joint = PersonUtil.Joint;
    using JKey = WearableAsset.JSONKey;

    /// <summary>
    /// Allows an asset to virtually parent itself directly to the joint of a person,
    /// rather than the control. This ensures that the asset will align with
    /// the position of the person's geometry. I made it for my hat :)
    /// </summary>
    public class WearableAsset : MVRScript
    {        
        public const string k_NONE = "None";
        public static class AtomType
        {
            public const string Person = "Person";
        }
        public static class Axis
        {
            public const int X = 0, Y = 1, Z = 2;
        }

        public static class JSONKey
        {
            public const string PoseEntry = "GVR_WearableAsset_Poses";
            public const string Name = "name";
            public const string Joint = "joint";
            public const string PersonIndex = "person"; // for backwards compat with old json, we aren't adding 'index' to the string.
            public const string PersonName = "person_name";

            public const string PosX = "pos_x";
            public const string PosY = "pos_y";
            public const string PosZ = "pos_z";

            public const string RotX = "rot_x";
            public const string RotY = "rot_y";
            public const string RotZ = "rot_z";
            public const string RotW = "rot_w";

            public const string OffPosX = "offpos_x";
            public const string OffPosY = "offpos_y";
            public const string OffPosZ = "offpos_z";

            public const string OffRotX = "offrot_x";
            public const string OffRotY = "offrot_y";
            public const string OffRotZ = "offrot_z";
        }

        private const string k_mainInstructs = "Poses are optional :)\n" +
                    "They store a snapshot of the transform and offset relative to a joint. " +
                    "You can also save or load them from other plugins like timeline. By default " +
                    "the pose will keep your current person selection, but if you start it with '@'" +
                    "it will target the person you saved it with (or the person at that index) \n\n" +
                    "<i>~~~BUTTONS~~~</i>\n" +
                    "<b>Commit Offset</b> = Merge the offset into the transformed position and reset. " +
                    "the object will stay put and your offsets will be 0. Probably only edge use cases.\n" +
                    "<b>Offset to Parent</b> = applies whatever offset is required to move the object " +
                    "to match the joint transform. You can then tweak the offsets. Theoretically a good path to keyframes.\n" +
                    "<b>Snap to Parent</b> = Clear offset and align the object with the joint.";

        public string Name => this.containingAtom != null ? this.containingAtom.name : "N/A";

        private JSONStorableStringChooser jointsTargetJSON;
        private JSONStorableStringChooser personTargetJSON;
        private JSONStorableBool debugModeJSON;
        private bool IsDebugMode => this.debugModeJSON != null && this.debugModeJSON.val;
        
        private JSONStorableBool persistOnChangeJSON;


        // todo: add scale offset & adoption

        private JSONStorableFloat positionXJSON;
        private JSONStorableFloat positionYJSON;
        private JSONStorableFloat positionZJSON;

        private JSONStorableFloat rotationXJSON;
        private JSONStorableFloat rotationYJSON;
        private JSONStorableFloat rotationZJSON;

        private UIDynamicSlider positionSliderX;
        private UIDynamicSlider positionSliderY;
        private UIDynamicSlider positionSliderZ;
        private UIDynamicSlider rotationSliderX;
        private UIDynamicSlider rotationSliderY;
        private UIDynamicSlider rotationSliderZ;
        private UIDynamic offsetSpacer;
        private int sliderIndex = -1;

        private JSONStorableAction commitOffsetActionJSON;
        private JSONStorableAction snapToParentActionJSON;
        private JSONStorableAction offsetToParentActionJSON;

        private JSONStorableString currentPoseNameJSON;
        private JSONStorableString savedPoseListJSON;
        private UIDynamicTextField savedPoseListRegion;
        private UIDynamicTextField currentPoseField;        

        private JSONStorableAction savePoseActionJSON;
        private JSONStorableAction loadPoseActionJSON;
        private JSONStorableAction deletePoseActionJSON;
        private UIDynamicButton savePoseButton;
        private UIDynamicButton loadPoseButton;
        private UIDynamicButton deletePoseButton;

        // these pose tools are registered for other plugins like timeline to use.
        // they are not displayed in the UI.
        private JSONStorableString storePoseJSON;
        private JSONStorableString loadPoseStringJSON;
        private JSONStorableStringChooser poseTargetsJSON;

        private JSONStorableBool allowCollisionJSON;

        private InputField savePoseNameField;
        

        private JSONStorableString instructionsJSON;
        private UIDynamicTextField instructionsField;

        private UIDynamicPopup jointPopup;

        private bool isSubscribed = false;
        private bool isInitialized = false;        
        private bool isInitializedUI = false;
        private bool hasPeople = false;

        private bool wasCollisionEnabled= false;
        private bool enableCollision = false;
        

        // we allow 1 frame of loading oddities before reporting pose errors.
        // this is because plugins in saved scenes may be restored before us,
        // causing a load attempt before we have generated poses.
        private bool hasLoaded = false;

        /// <summary>
        /// when loading a file, Init will populate our UI selections,
        /// including pose, before we actually load the poses.
        /// to handle this timing issue, we 
        /// </summary>
        private string initialPose = "";


        private string generalError = "";
        private bool everythingIsTerrible;
        private bool hasError => this.everythingIsTerrible || !string.IsNullOrEmpty(this.generalError);
        private bool hasValidJoint => this.personIndex >= 0 && !string.IsNullOrEmpty(this.jointsTargetJSON.val) && this.jointsTargetJSON.val != Joint.None;

        

        private List<string> joints = new List<string>()
        {
            Joint.None,
            Joint.Head, Joint.Neck,
            Joint.RShoulder, Joint.RUpperArm, Joint.RForearm, Joint.RHand, Joint.RThumb, Joint.RIndexF, Joint.RMiddleF, Joint.RRingF, Joint.RPinkyF,
            Joint.LShoulder, Joint.LUpperArm, Joint.LForearm, Joint.LHand, Joint.LThumb, Joint.LIndexF, Joint.LMiddleF, Joint.LRingF, Joint.LPinkyF,
            Joint.Chest, Joint.Abdomen, Joint.Hip,
            Joint.RThigh, Joint.RShin, Joint.RFoot,
            Joint.LThigh, Joint.LShin, Joint.LFoot
        };

        private List<Atom> people = new List<Atom>();
        private List<string> names = new List<string>();
        private int personIndex = -1;
        private int jointIndex = -1;
        private Transform targetJoint = null;
        private Transform personModelRoot = null;
        private Transform localTransform = null;
        private ParentTransformation transformation = new ParentTransformation();

        private Dictionary<string, Pose> poses = new Dictionary<string, Pose>();



        //private IEnumerator Start()
        //{
        //    this.hasLoaded = false;
        //    yield return null;
        //    this.hasLoaded = true;
        
        //}

        private void OnEnable()
        {
            this.DLog("OnEnable()");
            if (this.isInitialized && !this.isSubscribed)
            {                
                this.Subscribe();
                this.ScanForPeople();

                // for now we disable physics when activating the plugin.                
                var atom = this.GetContainingAtom();
                if (atom != null)
                {
                    atom.collisionEnabled = this.enableCollision;
                }
            }            
        }

        private void OnDisable()
        {
            this.DLog("OnDisable()");            
            this.Unsubscribe();

            // for now we disable physics when using the plugin.
            var atom = this.GetContainingAtom();
            if (atom != null)
            {
                this.enableCollision = atom.collisionEnabled;
                atom.collisionEnabled |= this.wasCollisionEnabled;
            }

        }


        /// <summary>
        /// When we have loaded from file, we want to load the stored poses we manually added to json.
        /// </summary>        
        public override void RestoreFromJSON(JSONClass json, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true)
        {
            this.DLog("RestoreFromJSON()");
            base.RestoreFromJSON(json, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);
            try
            {
                if (json.HasKey(JSONKey.PoseEntry))
                {
                    JSONArray poses = json[JSONKey.PoseEntry].AsArray;
                    foreach (var obj in poses)
                    {
                        Pose pose = new Pose(((JSONNode)obj).AsObject);
                        this.poses[pose.Name] = pose;
                    }
                    this.DLog("Poses built");

                }
            }
            catch (Exception e)
            {
                this.Log("WearableAsset failed to load poses from JSON.", isError: true);
            }
                        
            this.SyncPoseList();
            this.SyncInstructions();
            
        }

        /// <summary>
        /// Ensures any initial pose set by a plugin is in use,
        /// and that whatever the current offset values are they are 
        /// applied to the transformation
        /// </summary>
        public override void PostRestore()
        {
            this.DLog("PostRestore()");
            base.PostRestore();
            if (!string.IsNullOrEmpty(this.initialPose))
            {
                this.LoadPoseMethod(this.initialPose, external: false);
                this.initialPose = "";
            }
            this.SyncTransformation();
        }
        
      

        /// <summary>
        /// before we write to file, we convert our pose list to JSON and store it
        /// </summary>        
        public override JSONClass GetJSON(bool includeAppearance, bool includePhysical, bool forceStore)
        {
            var json = base.GetJSON(includeAppearance, includePhysical, forceStore);
            var array = new JSONArray();

            foreach (var keyValuePair in this.poses)
            {
                array.Add(keyValuePair.Value.GetJSON());
            }

            json[JSONKey.PoseEntry] = array;
            return json;
        }

     

        public override void Init()
        {
            this.DLog("Init()");


            base.Init();
            if (!this.isInitialized)
            {
                this.Unsubscribe();
                this.Subscribe();
                //this.GetNearbyBruteForce();
                this.isInitialized = true;
            }
            this.personTargetJSON = new JSONStorableStringChooser("SetPerson", new List<string>() { k_NONE }, k_NONE, "Target Person", this.OnSelectPerson);
            this.RegisterStringChooser(this.personTargetJSON);

            this.jointsTargetJSON = new JSONStorableStringChooser("SetJoint", this.joints, Joint.None, "Parent Joint", this.OnSelectJoint);
            this.RegisterStringChooser(this.jointsTargetJSON);
            
            this.poseTargetsJSON = new JSONStorableStringChooser("SelectPose", new List<string>() { k_NONE }, k_NONE, "Saved Poses", this.OnSelectPose);
            this.RegisterStringChooser(this.poseTargetsJSON);

            this.positionXJSON = new JSONStorableFloat("OffsetX", 0f, -0.5f, 0.5f, constrain:false);
            this.positionXJSON.setCallbackFunction += (value) => this.OnSetOffset(value, Axis.X, rotation:false, global:false);
            this.RegisterFloat(this.positionXJSON);

            this.positionYJSON = new JSONStorableFloat("OffsetY", 0f, -0.5f, 0.5f, constrain: false);
            this.positionYJSON.setCallbackFunction += (value) => this.OnSetOffset(value, Axis.Y, rotation: false, global: false);
            this.RegisterFloat(this.positionYJSON);

            this.positionZJSON = new JSONStorableFloat("OffsetZ", 0f, -0.5f, 0.5f, constrain: false);
            this.positionZJSON.setCallbackFunction += (value) => this.OnSetOffset(value, Axis.Z, rotation: false, global: false);
            this.RegisterFloat(this.positionZJSON);



            this.rotationXJSON = new JSONStorableFloat("RotationX", 0f, -360f, 360f, constrain: false);
            this.rotationXJSON.setCallbackFunction += (value) => this.OnSetOffset(value, Axis.X, rotation: true, global: false);
            this.RegisterFloat(this.rotationXJSON);

            this.rotationYJSON = new JSONStorableFloat("RotationY", 0f, -360f, 360f, constrain: false);
            this.rotationYJSON.setCallbackFunction += (value) => this.OnSetOffset(value, Axis.Y, rotation: true, global: false);
            this.RegisterFloat(this.rotationYJSON);

            this.rotationZJSON = new JSONStorableFloat("RotationZ", 0f, -360f, 360f, constrain: false);
            this.rotationZJSON.setCallbackFunction += (value) => this.OnSetOffset(value, Axis.Z, rotation: true, global: false);
            this.RegisterFloat(this.rotationZJSON);


            this.commitOffsetActionJSON = new JSONStorableAction("CommitOffset", this.CommitOffset);
            this.snapToParentActionJSON = new JSONStorableAction("SnapToParent", this.SnapToParent);           
            this.offsetToParentActionJSON = new JSONStorableAction("OffsetToParent", this.OffsetToParent);

            this.debugModeJSON = new JSONStorableBool("Debug", false);
            this.RegisterBool(this.debugModeJSON);

            this.localTransform = this.containingAtom.mainController.transform;


            this.currentPoseNameJSON = new JSONStorableString("CurrentPoseName", "Current Pose: None");
            this.savePoseActionJSON = new JSONStorableAction("SavePoseAction", this.OnClickSavePose);
            this.loadPoseActionJSON = new JSONStorableAction("LoadPoseAction", this.OnClickLoadPose);
            this.deletePoseActionJSON = new JSONStorableAction("DeletePoseAction", this.DeletePose);
            this.savedPoseListJSON = new JSONStorableString("SavedPoses","Saved Poses:");
            this.instructionsJSON = new JSONStorableString("Instructions", "Select a person to target");
            this.storePoseJSON = new JSONStorableString("StorePose", "", this.OnStorePose);
            this.RegisterString(this.storePoseJSON);
            this.loadPoseStringJSON = new JSONStorableString("PoseByName", "", this.OnSetPoseByName);
            this.RegisterString(this.loadPoseStringJSON);
            
            this.persistOnChangeJSON = new JSONStorableBool("Persist On Person Change", true);
            this.RegisterBool(this.persistOnChangeJSON);

            var atom = this.GetContainingAtom();
            if (atom != null)
            {
                this.wasCollisionEnabled = atom.collisionEnabled;
                this.enableCollision = false;
                atom.collisionEnabled = this.enableCollision;
            }
            //this.allowCollisionJSON = new JSONStorableBool("Allow Collision", false);
            //this.RegisterBool(this.allowCollisionJSON);

            
            this.ScanForPeople();
        }

        private void OnSetOffset(float value, int axis, bool rotation, bool global)
        {            
            if (rotation)
            {
                this.transformation.EulerOffset[axis] = value;
            }
            else
            {
                this.transformation.PositionOffset[axis] = value;
            }            
        }



        public override void InitUI()
        {
            this.DLog("init ui");
            base.InitUI();
            if (this.UITransform != null)
            {
                /// LEFT SIDE =======================================                
                // we store the joint popup because we'll be hiding it.
                // no need to cache the person popup, since it's always up.
                this.CreateScrollablePopup(this.personTargetJSON, false);
                this.jointPopup = this.CreateScrollablePopup(this.jointsTargetJSON);
                this.jointPopup.gameObject.SetActive(false);

                // there's a lot happening to generate pose saving elements,
                // so we handle that in its own method.
                this.GeneratePoseOptions();
                this.SetPoseOptionsVisible(false);

                this.instructionsField = this.CreateTextField(this.instructionsJSON, false);
                this.instructionsField.height = 300f;
                this.instructionsField.backgroundImage.color = Color.clear;
                this.instructionsField.UItext.fontStyle = FontStyle.Italic;

                /// RIGHT SIDE =======================================
                this.CreateToggle(this.debugModeJSON, true);
                this.CreateToggle(this.persistOnChangeJSON, true);
                //this.CreateToggle(this.allowCollisionJSON, true);

                var spacer = this.CreateSpacer(true);
                spacer.height = 20f;

                this.CreateOffsetSliders();

                spacer = this.CreateSpacer(true);
                spacer.height = 20f;

                var tempButton = this.CreateButton("Commit Offset", true);
                this.commitOffsetActionJSON.RegisterButton(tempButton);

                tempButton = this.CreateButton("Offset to Parent", true);
                this.offsetToParentActionJSON.RegisterButton(tempButton);

                tempButton = this.CreateButton("Snap to Parent", true);
                this.snapToParentActionJSON.RegisterButton(tempButton);
                
                this.isInitializedUI = true;

            }
        }


        private void Subscribe()
        {
            if (!this.isSubscribed)
            {                
                SuperController.singleton.onAtomUIDsChangedHandlers += this.OnAtomUIDsChanged;
                SuperController.singleton.onAtomAddedHandlers += this.OnAddAtom;
                SuperController.singleton.onAtomRemovedHandlers += this.OnRemoveAtom;
                this.containingAtom.mainController.onMovementHandlers += this.OnMoveAsset;
                this.isSubscribed = true;
            }
        }

        private void Unsubscribe()
        {
            if (this.containingAtom != null && this.containingAtom.mainController != null)
            {
                this.containingAtom.mainController.onMovementHandlers -= this.OnMoveAsset;
            }

            if (SuperController.singleton != null)
            {
                SuperController.singleton.onAtomUIDsChangedHandlers -= this.OnAtomUIDsChanged;
                SuperController.singleton.onAtomAddedHandlers -= this.OnAddAtom;
                SuperController.singleton.onAtomRemovedHandlers -= this.OnRemoveAtom;                            
            }
            this.isSubscribed = false;
        }

        private void OnMoveAsset(FreeControllerV3 controller)
        {
            if (this.targetJoint != null)
            {
                this.transformation.Store(controller.transform, targetJoint);
            }
        }

        private void CreateOffsetSliders()
        {
            if (this.positionSliderX != null) this.RemoveSlider(this.positionSliderX);
            if (this.positionSliderY != null) this.RemoveSlider(this.positionSliderY);
            if (this.positionSliderZ != null) this.RemoveSlider(this.positionSliderZ);
            if (this.rotationSliderX != null) this.RemoveSlider(this.rotationSliderX);
            if (this.rotationSliderY != null) this.RemoveSlider(this.rotationSliderY);
            if (this.rotationSliderZ != null) this.RemoveSlider(this.rotationSliderZ);

            Debug.Log("creating pos");
            this.positionSliderX = this.CreateSlider(this.positionXJSON, true);
            if (this.sliderIndex < 0f) this.sliderIndex = this.positionSliderX.transform.GetSiblingIndex();
            this.positionSliderY = this.CreateSlider(this.positionYJSON, true);
            this.positionSliderZ = this.CreateSlider(this.positionZJSON, true);

            Debug.Log("creating spacer");
            if (this.offsetSpacer == null)
            {
                this.offsetSpacer = this.CreateSpacer(true);
                this.offsetSpacer.height = 20f;
            }

            Debug.Log("creating rot");
            this.rotationSliderX = this.CreateSlider(this.rotationXJSON, true);
            this.rotationSliderY = this.CreateSlider(this.rotationYJSON, true);
            this.rotationSliderZ = this.CreateSlider(this.rotationZJSON, true);

            Debug.Log("assigning indices");
            if (this.sliderIndex >= 0)
            {
                Debug.Log("pos ");
                this.positionSliderX.transform.SetSiblingIndex(this.sliderIndex);
                this.positionSliderY.transform.SetSiblingIndex(this.sliderIndex+1);
                this.positionSliderZ.transform.SetSiblingIndex(this.sliderIndex+2);
                Debug.Log("offset");
                this.offsetSpacer.transform.SetSiblingIndex(this.sliderIndex+3);
                Debug.Log("rot ");
                this.rotationSliderX.transform.SetSiblingIndex(this.sliderIndex+4);
                this.rotationSliderY.transform.SetSiblingIndex(this.sliderIndex+5);
                this.rotationSliderZ.transform.SetSiblingIndex(this.sliderIndex+6);

            }
            Debug.Log("done");
        }
        


        private void SyncPoseList()
        {
            List<string> poseNames = new List<string>(this.poses.Keys);
            poseNames.Sort();
            System.Text.StringBuilder sb = new System.Text.StringBuilder();
            sb.AppendLine("<i>Saved Poses:</i>");
            foreach(var poseName in poseNames)
            {
                sb.AppendLine(poseName);
            }
            this.savedPoseListJSON.val = sb.ToString();
            poseNames.Insert(0, k_NONE);
            this.poseTargetsJSON.choices = poseNames;
        }

        private void SyncInstructions()
        {
            if (this.personIndex < 0)
            {
                this.instructionsJSON.val = "Select a person above to target";
            }
            else if (this.targetJoint == null)
            {
                this.instructionsJSON.val = "Select a joint to parent to";
            }
            else
            {
                this.instructionsJSON.val = k_mainInstructs;
            }
        }


        /// <summary>
        /// Re-iterates our known people to ensure our names are up to date.
        /// </summary>        
        private void OnAtomUIDsChanged(List<string> uids)
        {
            this.RefreshPersonNames();
        }


        /// <summary>
        /// Sets the root person and hides or shows the joint display.
        /// Called when a selection is made in the person list JSON
        /// </summary>
        private void OnSelectPerson(string uid)
        {            
            this.DLog($"OnSelectPerson: {uid}");
            if (!string.IsNullOrEmpty(uid))
            {
                int index = -1;
                if (uid != k_NONE) index = this.names.IndexOf(uid);
                this.SelectPersonMethod(index);
            }

        }
        

        private void SelectPersonMethod(int index)
        {
            int previousIndex = this.personIndex;
            this.personIndex = index;
            if (this.personIndex < 0 || this.personIndex >= this.people.Count || this.people[personIndex] == null)
            {
                // if we select NONE or are deselecting via -1 index or select an invalid range or missing atom,
                // we treat all of these as a deselection.
                this.personIndex = -1;
                this.personModelRoot = null;
                this.personTargetJSON.valNoCallback = k_NONE;
                this.jointsTargetJSON.valNoCallback = Joint.None;
                this.targetJoint = null;
                this.SetJointPopupVisible(false);
                this.SetPoseOptionsVisible(false);
                this.DLog($"disconnected from person (None selected or person not valid)");
            }
            else
            {
                // otherwise we update our root and enable the joint selector if necessary.
                Atom newPerson = this.people[this.personIndex];
                this.DLog($"loading root of '{newPerson.name}'");
                this.personModelRoot = PersonUtil.FindRoot(newPerson);
                if (previousIndex < 0)
                {
                    // if we didn't have a person selected before, we will need to show the joints.
                    this.SetJointPopupVisible(true);
                    this.DLog($"now connected to '{this.people[this.personIndex].uid}.'");
                }
                else
                {
                    // if we did, we should synchronize the joint to the new person instead,
                    // either keeping its current position or not.
                    if (this.persistOnChangeJSON.val) this.SelectJointMethod(this.jointsTargetJSON.val, clearPose:false);
                    else this.SelectJointNoStore(this.jointsTargetJSON.val, clearPose:false);

                    if (previousIndex < this.people.Count && previousIndex != this.personIndex)
                    {
                        this.DLog($"now connected to '{this.people[this.personIndex].uid}' instead of '{this.people[previousIndex].uid}'.");
                    }
                    else
                    {
                        this.DLog($"now connected to '{this.people[this.personIndex].uid}.'");
                    }
                }

                if (this.personModelRoot == null) this.Log("Failed to find person model root, will not be able to find joints.", isError: true);
            }

            this.SyncInstructions();
        }

        // need to change to select by index, and allow a json where timeline
        // can apply a selection by index in case names change.




        /// <summary>
        /// Selects a joint according to the name, clearing any current pose from the UI.
        /// this is called when the joint string chooser value changes, either from a plugin
        /// or from interacting with the UI.
        /// </summary>
        /// <param name="jointName"></param>
        private void OnSelectJoint(string jointName) => this.SelectJointMethod(jointName, clearPose: true);
        
        /// <summary>
        /// Finds and targets the specified joint by id, if we have a root model.
        /// Called when the joint select list changes, or when we swap people with a joint selected.
        /// </summary>
        /// <param name="jointName"></param>
        /// <param name="clearPose">if true, the current pose will be cleared. False when we target a joint due to loading a pose itself.</param>
        private void SelectJointMethod(string jointName, bool clearPose = true)
        {
            this.DLog($"OnSelectJoint({jointName})");
            if (this.personModelRoot != null && !string.IsNullOrEmpty(jointName))
            {
                this.targetJoint = null;
                if (jointName != Joint.None)
                {
                    this.targetJoint = PersonUtil.FindJoint(this.personModelRoot, jointName);
                }

                if (this.targetJoint != null)
                {
                    this.jointIndex = this.joints.IndexOf(jointName);
                    this.transformation.Store(this.localTransform, this.targetJoint);
                    this.DLog($"now connected to joint: '{jointName}'.");
                    if (clearPose)
                    {
                        this.currentPoseNameJSON.val = "Current Pose: None";
                        this.poseTargetsJSON.val = k_NONE;
                        this.savePoseNameField.text = "";                        
                    }
                    this.SetPoseOptionsVisible(true);
                }
                else
                {
                    this.SetPoseOptionsVisible(false);
                    this.LogIf(jointName != Joint.None, $"failed to find joint '{jointName}'", isError: true);
                }
            }
            this.SyncInstructions();
        }

        /// <summary>
        /// Essentially the same as the OnSelectJoint callback,
        /// but we don't store the current position into the offset.
        /// This causes the object to snap to the new joint using its current offset.
        /// We only use this when our persist mode is false when changing a new person,
        /// and it allows easily swapping who is wearing the item.
        /// </summary>
        /// <param name="jointName">the name of joint to find and target</param>
        /// <param name="clearPose">if true, the current pose will be cleared. False when we target a joint due to loading a pose itself.</param>
        private void SelectJointNoStore(string jointName, bool clearPose = true)
        {
            if (this.personModelRoot != null && !string.IsNullOrEmpty(jointName))
            {
                this.targetJoint = null;
                if (jointName != Joint.None)
                {
                    this.targetJoint = PersonUtil.FindJoint(this.personModelRoot, jointName);
                }

                if (this.targetJoint != null)
                {
                    this.DLog($"now connected to joint: '{jointName}'.");

                    if (clearPose)
                    {
                        this.currentPoseNameJSON.val = "Current Pose: None";
                        this.poseTargetsJSON.val = k_NONE;
                        this.savePoseNameField.text = "";
                    }
                    this.SetPoseOptionsVisible(true);
                }
                else
                {
                    this.SetPoseOptionsVisible(false);
                    this.LogIf(jointName != Joint.None, $"failed to find joint '{jointName}'", isError: true);
                }
            }
            this.SyncInstructions();
        }

        private void SyncPoseUI()
        {
            string current = this.savePoseNameField.text;

            if (string.IsNullOrEmpty(current))
            {
                this.savePoseButton.buttonColor = Color.white;
                this.savePoseButton.label = "Save Pose";
                this.savePoseButton.button.interactable = false;
                this.loadPoseButton.button.interactable = false;
                this.deletePoseButton.button.interactable = false;
            }
            else if (this.poses.ContainsKey(current))
            {
                this.savePoseButton.buttonColor = Color.yellow;
                this.savePoseButton.label = "Overwrite Pose";
                this.savePoseButton.button.interactable = true;
                this.loadPoseButton.button.interactable = true;
                this.deletePoseButton.button.interactable = true;
            }
            else
            {
                this.savePoseButton.buttonColor = Color.white;
                this.savePoseButton.label = "Save Pose";
                this.savePoseButton.button.interactable = true;
                this.loadPoseButton.button.interactable = false;
                this.deletePoseButton.button.interactable = false;
            }
        }

        /// <summary>
        /// Refreshes the display for person and joint string choosers,
        /// in the case that they were indirectly changed such as when loading a pose.
        /// this does not handle any actual targeting or index setting, only display.
        /// as such if targeting is not properly handled elsewhere, the display will be faulty.
        /// </summary>
        private void SyncChooserUI()
        {
            if (this.personIndex >= 0 && this.personIndex < this.people.Count)
            {
                this.personTargetJSON.valNoCallback = this.people[this.personIndex].name;
                this.SetJointPopupVisible(true);
            }
            else
            {                
                this.personTargetJSON.valNoCallback = k_NONE;
                this.SetJointPopupVisible(false);
            }


            if (this.jointIndex >= 0 && this.jointIndex < this.joints.Count)
            {
                this.jointsTargetJSON.valNoCallback = this.joints[this.jointIndex];
                this.SetPoseOptionsVisible(true);
            }
            else
            {                
                this.jointsTargetJSON.valNoCallback = Joint.None;
                this.SetPoseOptionsVisible(false);
            }
        }

        /// <summary>
        /// Applies the offset values to the transformation object,
        /// as the initialization may have loaded them before a joint was available.
        /// </summary>
        private void SyncTransformation()
        {

        }
        
        private void OnSetPoseNameField(string value)
        {            
            this.SyncPoseUI();
        }

        private void OnStorePose(string poseName)
        {
            this.SavePoseMethod(poseName);
            this.storePoseJSON.valNoCallback = "";
        }

        private void OnClickSavePose() => this.SavePoseMethod(this.savePoseNameField.text);
        private void SavePoseMethod(string poseName)
        {
            if (!string.IsNullOrEmpty(poseName))
            {
                Pose pose;
                if (!this.poses.TryGetValue(poseName, out pose))
                {
                    pose = new Pose();
                    this.poses[poseName] = pose;
                }

                pose.Name = poseName;
                pose.PersonName = this.personIndex >= 0 && this.personIndex < this.people.Count && this.people[personIndex] != null ? this.people[personIndex].name : "";
                pose.FallbackPersonIndex = this.personIndex;
                pose.Joint = this.jointsTargetJSON.val;
                pose.Position = this.transformation.Position;
                pose.Rotation = this.transformation.Rotation;
                pose.PositionOffset = this.transformation.PositionOffset;
                pose.EulerOffset = this.transformation.EulerOffset;

                this.Log($"pose '{poseName}' saved.");
                this.SyncPoseList();
                this.SyncPoseUI();
            }
        }

        private void DeletePose() => this.DeletePoseMethod(this.savePoseNameField.text);

        private void DeletePoseMethod(string poseName)
        {
            if (!string.IsNullOrEmpty(poseName))
            {
                this.poses.Remove(poseName);
                this.SyncPoseList();
                this.SyncPoseUI();
            }
        }

        /// <summary>
        /// The load call when selecting a pose from the hidden
        /// 'chooser' json, such as from timeline.
        /// </summary>        
        private void OnSelectPose(string value)
        {
            this.DLog($"OnSelectPose({value})");
            this.LoadPoseMethod(value, external: true);
        } 

        /// <summary>
        /// the load call when clicking the load pose button in the ui
        /// </summary>        
        private void OnClickLoadPose()
        {
            this.DLog($"OnClickLoadPose()");
            this.LoadPoseMethod(this.savePoseNameField.text, external:false);
        }

        /// <summary>
        /// the load call when a plugin changes the PoseByName pose string.
        /// this effect is exposed but not a visible UI element
        /// </summary>
        /// <param name="posename"></param>
        private void OnSetPoseByName(string posename)
        {
            this.DLog($"OnLoadPoseExternal({posename})");
            this.LoadPoseMethod(posename, external:true);
            this.loadPoseStringJSON.valNoCallback = "";
        }

        /// <summary>
        /// The actual logic behind all pose loading.
        /// </summary>        
        private void LoadPoseMethod(string poseName, bool external)
        {
            this.DLog($"LoadPoseMethod({poseName})");
            
            Pose pose = null;
            if (!string.IsNullOrEmpty(poseName) && poseName != k_NONE)
            {
                bool includeTarget = this.personIndex < 0 || poseName.StartsWith("@");
                if (poseName.StartsWith("@")) poseName = poseName.Substring(1);

                if (this.poses.TryGetValue(poseName, out pose))
                {
                    if (includeTarget)
                    {
                        Atom person = null;
                        if (!string.IsNullOrEmpty(pose.PersonName))
                        {
                            // since I'm not 10000% sure the people will always be
                            // loaded into the atom list in order, on loading a scene json,
                            // we try to match the person by name first when loading a pose.
                            for (int i = 0; i < this.people.Count; i++)
                            {
                                if (this.people[i].name == pose.PersonName)
                                {
                                    this.personIndex = i;
                                    break;
                                }
                            }
                        }
                        else
                        {
                            // if we can't, then we use the saved index.
                            this.personIndex = pose.FallbackPersonIndex;
                        }

                        if (this.personIndex >= 0 && this.personIndex < this.people.Count)
                        {                            
                            this.OnSelectPerson(this.names[this.personIndex]);
                        }
                        else if (this.people.Count > 0)
                        {
                            this.Log($"couldn't load the person target for pose '{pose.Name}', the index was invalid. " +
                                $"index is {this.personIndex}, valid range is 0-{this.people.Count}", isError: true);
                        }
                        else
                        {
                            this.Log($"couldn't load the targets for pose '{pose.Name}', because there are no people.", isError: true);
                        }
                    }

                    if (this.personIndex >= 0)
                    {
                        this.SelectJointMethod(pose.Joint, clearPose: false);                        
                    }

                    if (this.targetJoint != null)
                    {
                        this.transformation.Position = pose.Position;
                        this.transformation.Rotation = pose.Rotation;
                        this.transformation.PositionOffset = pose.PositionOffset;
                        this.transformation.EulerOffset = pose.EulerOffset;
                        this.SyncOffset();
                    }
                    else
                    {                        
                        this.Log($"couldn't load the pose '{pose.Name}'.", isError: true);
                    }
                }
                else if (!this.hasLoaded && external)
                {
                    // when loading scenes, other plugins like timeline might
                    // restore before us and try to load a pose before we've loaded it.
                    // so we cache the 'reqeuest' for a frame to give ourselves a chance to load poses.
                    this.initialPose = poseName;
                }
                else
                {                    
                    this.Log($"no pose named '{poseName}'", isError: true);
                }
            }

            
            if (pose != null)
            {
                
                this.savePoseNameField.text = pose.Name;
                this.currentPoseNameJSON.val = "Current Pose: " + pose.Name;
                
            }
            else
            {
                this.savePoseNameField.text = "";
                this.currentPoseNameJSON.val = "Current Pose: None";
            }

            this.SyncChooserUI();
            this.SyncPoseUI();

        }

        private void SyncOffset()
        {
            this.positionXJSON.valNoCallback = this.transformation.PositionOffset.x;
            this.positionYJSON.valNoCallback = this.transformation.PositionOffset.y;
            this.positionZJSON.valNoCallback = this.transformation.PositionOffset.z;

            this.rotationXJSON.valNoCallback = this.transformation.EulerOffset.x;
            this.rotationYJSON.valNoCallback = this.transformation.EulerOffset.y;
            this.rotationZJSON.valNoCallback = this.transformation.EulerOffset.z;

        }

        private void SetPoseOptionsVisible(bool visible)
        {
            this.currentPoseField.gameObject.SetActive(visible);
            this.savePoseButton.gameObject.SetActive(visible);
            this.loadPoseButton.gameObject.SetActive(visible);
            this.deletePoseButton.gameObject.SetActive(visible);
            this.savePoseNameField.gameObject.SetActive(visible);
            this.savedPoseListRegion.gameObject.SetActive(visible);
        }

        /// <summary>
        /// This builds all the UI elements responsible for saving poses,
        /// called during InitUI()
        /// </summary>
        private void GeneratePoseOptions()
        {
            UIDynamic spacerTemp = this.CreateSpacer(false);
            spacerTemp.height = 30f;

            LayoutElement elementTemp;

            // we want to display the current pose, but we have to tweak the text field
            // because it likes to be big and in the way, and we need it smaller.
            this.currentPoseField = this.CreateTextField(this.currentPoseNameJSON, false);            
            this.currentPoseField.UItext.alignment = TextAnchor.MiddleCenter;
            this.currentPoseField.UItext.fontStyle = FontStyle.BoldAndItalic;
            this.currentPoseField.backgroundImage.color = Color.clear;
            elementTemp = this.currentPoseField.GetComponent<LayoutElement>();
            elementTemp.minHeight = 40f;
            elementTemp.preferredHeight = 40f;

            
            // next we need the save and load buttons
            this.savePoseButton = this.CreateButton("Save Pose", false);            
            this.savePoseActionJSON.RegisterButton(this.savePoseButton);
            this.savePoseButton.button.interactable = false;
            this.loadPoseButton = this.CreateButton("Load Pose", false);
            this.loadPoseActionJSON.RegisterButton(this.loadPoseButton);            
            this.loadPoseButton.button.interactable = false;
            this.deletePoseButton = this.CreateButton("Delete Pose", false);
            this.deletePoseActionJSON.RegisterButton(this.deletePoseButton);
            this.deletePoseButton.buttonColor = Color.red;
            this.deletePoseButton.textColor = Color.white;
            this.deletePoseButton.button.interactable = false;

            // we want to display a list of all poses saved
            this.savedPoseListRegion = this.CreateTextField(this.savedPoseListJSON, false);
            this.savedPoseListRegion.height = 300f;
            
            // we want to use an input field to let people save custom 'poses' of assets.
            // there doesn't seem to be an integrated way to create an input field.
            // I don't feel like building a factory for it, so I'm just going to duplicate 
            // an existing one I know about: the label all plugins inherently use for the plugin manager
            // edit: in hindsight this might actually be more lines of code than using a factory. meh.
            if (this.pluginLabelJSON.inputField != null)
            {
                this.savePoseNameField = Instantiate(this.pluginLabelJSON.inputField, this.savePoseButton.transform.parent);
                elementTemp = this.savePoseNameField.gameObject.AddComponent<LayoutElement>();
                elementTemp.minHeight = 60f;
                elementTemp.flexibleWidth = 100f;
                this.savePoseNameField.transform.Find("Placeholder").GetComponent<Text>().text = "Enter Pose Name...";
                this.savePoseNameField.name = "SavePoseInputField";
                this.savePoseNameField.transform.SetSiblingIndex(this.savePoseButton.transform.GetSiblingIndex());
                this.savePoseNameField.onValueChanged.RemoveAllListeners();
                this.savePoseNameField.onEndEdit.RemoveAllListeners();
                this.savePoseNameField.onValueChanged.AddListener(this.OnSetPoseNameField);
            }
            else
            {
                this.Log("unable to create input field for pose name, couldn't find a prefab.", isError: true);
            }

        }

        /// <summary>
        /// Applies the offset of the virtual parent to the 'actual' 
        /// virtual child position and rotation by combining them and 
        /// setting the offsets to zero.
        /// </summary>
        /// <remarks>
        /// be careful if using this in tandem with timeline, as timeline
        /// will 'unclear' the offset if its currently being animated. as such,
        /// you will either need to set the offset to zero in timeline or time
        /// the animation to be transient and start the next one at zero.
        /// </remarks>
        public void CommitOffset()
        {
            this.transformation.CommitOffset();
            this.ClearOffsetUI();
        }


        /// <summary>
        /// Calculates the difference in position and euler angles required 
        /// to exactly match the parent joint, and adds that into
        /// the current offset. This might useful if you want to animate an object
        /// moving to the joint root position, but that needs to be tested
        /// </summary>
        public void OffsetToParent()
        {
            if (this.targetJoint != null)
            {                               
                Vector3 posOffset = -this.transformation.Position;
                Vector3 eulerOffset = Quaternion.Inverse(this.transformation.Rotation).eulerAngles;
                this.positionXJSON.val = posOffset.x;
                this.positionYJSON.val = posOffset.y;
                this.positionZJSON.val = posOffset.z;
                this.rotationXJSON.val = eulerOffset.x;
                this.rotationYJSON.val = eulerOffset.y;
                this.rotationZJSON.val = eulerOffset.z;                
            }
        }


        /// <summary>
        /// Clears all offsets and sets the localized position and rotation
        /// to zero
        /// </summary>
        public void SnapToParent()
        {
            this.ClearOffsetUI();
            this.transformation.SetZero();
        }

        private void ClearOffsetUI()
        {
            this.positionXJSON.val = 0f;
            this.positionYJSON.val = 0f;
            this.positionZJSON.val = 0f;
            this.rotationXJSON.val = 0f;
            this.rotationYJSON.val = 0f;
            this.rotationZJSON.val = 0f;

        }


        private void SetJointPopupVisible(bool visible)
        {
            // check null to account for destruction timing in quit and scene close,
            // as we use this method when removing a target, and they might be
            // removed after our UI is destroyed but before we are destroyed.
            if (this.jointPopup != null && this.jointPopup.gameObject != null)
            {
                this.jointPopup.gameObject.SetActive(visible);
                if (!visible) this.jointsTargetJSON.val = Joint.None;
            }
        }

        /// <summary>
        /// Searches current scene atoms for people, and populates the list of persons
        /// </summary>
        private void ScanForPeople()
        {
            this.DLog("ScanForPeople");
            foreach (var atom in this.GetSceneAtoms())
            {
                if (atom.type == AtomType.Person && !this.people.Contains(atom))
                {
                    this.DLog("Adding Person "+atom.uid);
                    this.people.Add(atom);
                }
            }

            this.RefreshPersonNames();
        }

        /// <summary>
        /// Adds any newly added people to the person list,
        /// called via subscription to SuperCOntroller.addAtomCallbackHandlers
        /// </summary>        
        private void OnAddAtom(Atom atom)
        {            
            if (atom.type == AtomType.Person && !this.people.Contains(atom))
            {
                this.DLog("Adding Person " + atom.uid);
                this.people.Add(atom);
                this.RefreshPersonNames();
            }
        }

        /// <summary>
        /// Checks to see if we're removing our current target,
        /// and if so, clears our target. Called when an atom is removed
        /// via subscription to SuperController.onRemoveAtomHandlers
        /// </summary>
        /// <param name="atom"></param>
        private void OnRemoveAtom(Atom atom)
        {
            if (atom != null && atom.type == AtomType.Person)
            {
                int index = this.people.IndexOf(atom);
                if (index >= 0)
                {
                    this.people.Remove(atom);
                    this.names.RemoveAt(index);
                    if (index == this.personIndex)
                    {
                        this.targetJoint = null;
                        this.personTargetJSON.val = k_NONE;
                        this.jointsTargetJSON.val = k_NONE;                        
                    }
                    this.RefreshPersonNames();
                }
            }
        }

        private void RefreshPersonNames()
        {
            this.DLog("RefreshPersonNames 1234");
            if (this.personTargetJSON != null)
            {
                this.names.Clear();
                foreach (var person in this.people)
                {
                    if (person != null)
                    { 
                        this.names.Add(person.uid);
                        this.DLog("found person: " + person.uid);
                    }
                }
                
                List<string> choices = new List<string>(this.names);
                choices.Insert(0, k_NONE);
                this.personTargetJSON.choices = choices;
            }
        }

       

        private void LateUpdate()
        {
            
            if (this.personIndex >= 0 && this.targetJoint != null)
            {
                this.transformation.Apply(this.localTransform, this.targetJoint);
                
            }
            
            
        }

        private void DLog(string message, bool isError = false)
        {
            this.LogIf(this.IsDebugMode, message, isError);
        }

        private void LogIf(bool condition, string message, bool isError = false)
        {
            if (condition) this.Log(message, isError);
        }
        private void Log(string msg, bool isError = false)
        {
            msg = $"WearableAsset '{this.Name}': " + msg;
            if (!isError) SuperController.LogMessage(msg);
            else SuperController.LogError(msg);
        }

        
        private class Pose
        {
            //public bool IsValid { get; private set; }
            
            public string Name;

            /// <summary>
            /// The name of the person we're targeting.
            /// We will prefer this when trying to load a pose.
            /// </summary>
            public string PersonName;

            /// <summary>
            /// The index of the person in addition to the name
            /// is stored in the pose. We try to find our target by name,
            /// but if we can't, we use the person that is at their
            /// index in the list.
            /// </summary>
            public int FallbackPersonIndex;
            public string Joint;
            public Vector3 Position;
            public Quaternion Rotation;
            public Vector3 PositionOffset;
            public Vector3 EulerOffset;

            public Pose() { }

            public Pose(JSONClass json)
            {
                try
                {
                    this.Name = json[JKey.Name].Value;
                               
                    this.FallbackPersonIndex = json[JKey.PersonIndex].AsInt;
                    this.PersonName = json[JKey.PersonName];
                    this.Joint = json[JKey.Joint].Value;
                    
                    this.Position.x = json[JKey.PosX].AsFloat;
                    this.Position.y = json[JKey.PosY].AsFloat;
                    this.Position.z = json[JKey.PosZ].AsFloat;

                    this.Rotation.x = json[JKey.RotX].AsFloat;
                    this.Rotation.y = json[JKey.RotY].AsFloat;
                    this.Rotation.z = json[JKey.RotZ].AsFloat;
                    this.Rotation.w = json[JKey.RotW].AsFloat;

                    this.PositionOffset.x = json[JKey.OffPosX].AsFloat;
                    this.PositionOffset.y = json[JKey.OffPosY].AsFloat;
                    this.PositionOffset.z = json[JKey.OffPosY].AsFloat;

                    this.EulerOffset.x = json[JKey.OffRotX].AsFloat;
                    this.EulerOffset.y = json[JKey.OffRotY].AsFloat;
                    this.EulerOffset.z = json[JKey.OffRotZ].AsFloat;
            
                }
                catch(Exception e)
                {
                    SuperController.LogError("WearableAsset exception building pose from JSON: " + e);
                }
            }


            public JSONClass GetJSON()
            {
                var ret_json = new JSONClass();
                ret_json[JKey.Name] = this.Name;
                ret_json[JKey.PersonIndex] = this.FallbackPersonIndex.ToString();
                ret_json[JKey.PersonName] = this.PersonName;
                ret_json[JKey.Joint] = this.Joint;
                
                ret_json[JKey.PosX] = this.Position.x.ToString();
                ret_json[JKey.PosY] = this.Position.y.ToString();
                ret_json[JKey.PosZ] = this.Position.z.ToString();

                ret_json[JKey.RotX] = this.Rotation.x.ToString();
                ret_json[JKey.RotY] = this.Rotation.y.ToString();
                ret_json[JKey.RotZ] = this.Rotation.z.ToString();
                ret_json[JKey.RotW] = this.Rotation.w.ToString();

                ret_json[JKey.OffPosX] = this.PositionOffset.x.ToString();
                ret_json[JKey.OffPosY] = this.PositionOffset.y.ToString();
                ret_json[JKey.OffPosZ] = this.PositionOffset.z.ToString();

                ret_json[JKey.OffRotX] = this.EulerOffset.x.ToString();
                ret_json[JKey.OffRotY] = this.EulerOffset.y.ToString();
                ret_json[JKey.OffRotZ] = this.EulerOffset.z.ToString();
                return ret_json;
            }
        }
    }
}