namespace GossamerVR.WearableAsset
{
    using UnityEngine;

    /// <summary>
    /// Stores and retrieves a position and rotation
    /// offset from one transform to another.
    /// </summary>
    public class ParentTransformation
    {
        // these are additional offset values that don't get stored,
        // in fact they get accounted for and removed from storage when
        // Store() is called, as they are expected to be animated values
        // that change between Apply() calls.
        public Vector3 PositionOffset;
        public Vector3 EulerOffset;
        
        public Vector3 Position;
        public Quaternion Rotation;

        public void Store(Transform child, Transform parent)
        {
            this.Position = parent.InverseTransformPoint(child.position) - this.PositionOffset;
            this.Rotation = Quaternion.Inverse(parent.rotation) * child.rotation * Quaternion.Inverse(Quaternion.Euler(this.EulerOffset));
            
        }

        /// <summary>
        /// sets the position and rotation of the 'child' transform
        /// based on the stored offset, relative to the 'parent' transform.
        /// </summary>        
        public void Apply(Transform child, Transform parent)
        {
            child.position = parent.TransformPoint(this.Position + this.PositionOffset);
            child.rotation = parent.rotation * this.Rotation * Quaternion.Euler(this.EulerOffset);
        }

        /// <summary>
        /// Stores the current extra offset as part of the permanent offset.
        /// Usually done when changing parents or snapping to a parent        
        /// </summary>
        public void CommitOffset()
        {
            this.Position += this.PositionOffset;
            this.PositionOffset = Vector3.zero;
            this.Rotation *= Quaternion.Euler(this.EulerOffset);
            this.EulerOffset = Vector3.zero;           
        }

        /// <summary>
        /// clears the stored relative position and rotation,
        /// but not the stored offset.
        /// </summary>
        public void SetZero()
        {
            this.Position = Vector3.zero;
            this.Rotation = Quaternion.identity;
        }

    }
}
